from builtins import str
from collections import namedtuple
from functools import wraps

from parsimonious.exceptions import ParseError as ParsimoniousParseError

from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context, Evaluator
from pyjexl.exceptions import ParseError
from pyjexl.operators import default_binary_operators, default_unary_operators, Operator
from pyjexl.parser import jexl_grammar, Parser


#: Encapsulates the variable parts of JEXL that affect parsing and
#: evaluation.
JEXLConfig = namedtuple(
    "JEXLConfig", ["functions", "transforms", "unary_operators", "binary_operators"]
)


def invalidates_grammar(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self._grammar = None
        return func(self, *args, **kwargs)

    return wrapper


class JEXL(object):
    def __init__(self, context=None):
        self.context = Context(context or {})
        self.config = JEXLConfig(
            functions={},
            transforms={},
            unary_operators=default_unary_operators.copy(),
            binary_operators=default_binary_operators.copy(),
        )

        self._grammar = None

    @property
    def grammar(self):
        if not self._grammar:
            self._grammar = jexl_grammar(self.config)
        return self._grammar

    @invalidates_grammar
    def add_binary_operator(self, operator, precedence, func):
        self.config.binary_operators[operator] = Operator(operator, precedence, func)

    @invalidates_grammar
    def remove_binary_operator(self, operator):
        del self.config.binary_operators[operator]

    @invalidates_grammar
    def add_unary_operator(self, operator, func):
        self.config.unary_operators[operator] = Operator(operator, 1000, func)

    @invalidates_grammar
    def remove_unary_operator(self, operator):
        del self.config.unary_operators[operator]

    def add_transform(self, name, func):
        self.config.transforms[name] = func

    def remove_transform(self, name):
        del self.config.transforms[name]

    def transform(self, name=None):
        def wrapper(func):
            self.config.transforms[name or func.__name__] = func
            return func

        return wrapper

    def add_function(self, name, func):
        self.config.functions[name] = func

    def remove_functions(self, name):
        del self.config.functions[name]

    def function(self, name=None):
        def wrapper(func):
            self.config.functions[name or func.__name__] = func
            return func

        return wrapper

    def parse(self, expression):
        try:
            return Parser(self.config).visit(self.grammar.parse(expression))
        except ParsimoniousParseError:
            raise ParseError("Could not parse expression: " + expression)

    def analyze(self, expression, AnalyzerClass):
        parsed_expression = self.parse(expression)
        visitor = AnalyzerClass(self.config)
        return visitor.visit(parsed_expression)

    def validate(self, expression):
        try:
            for res in self.analyze(expression, ValidatingAnalyzer):
                yield res
        except ParseError as err:
            yield str(err)

    def evaluate(self, expression, context=None):
        # If expression is a string, parse it; otherwise, assume it's already parsed
        if isinstance(expression, str):
            parsed_expression = self.parse(expression)
        else:
            parsed_expression = expression
        context = Context(context) if context is not None else self.context
        return Evaluator(self.config).evaluate(parsed_expression, context)



import pytest
from pyjexl.jexl import JEXL

def test_evaluate_with_parsed_expression():
    jexl = JEXL()
    expr = '2 + 3'
    parsed = jexl.parse(expr)
    # Should work with parsed expression
    assert jexl.evaluate(parsed) == 5
    # Should work with string as before
    assert jexl.evaluate(expr) == 5



from collections.abc import MutableMapping
from pyjexl.exceptions import MissingFunctionError, MissingTransformError


class Context(MutableMapping):
    def __init__(self, context_data=None):
        self.data = context_data or {}
        self.relative_value = {}

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def with_relative(self, relative_value):
        new_context = Context(self.data)
        new_context.relative_value = relative_value
        return new_context


class Evaluator(object):
    def __init__(self, jexl_config):
        self.config = jexl_config

    def evaluate(self, expression, context=None):
        method = getattr(self, "visit_" + type(expression).__name__, self.generic_visit)
        context = context or Context()
        return method(expression, context)

    def visit_BinaryExpression(self, exp, context):
        return exp.operator.do_evaluate(
            lambda: self.evaluate(exp.left, context),
            lambda: self.evaluate(exp.right, context),
        )

    def visit_UnaryExpression(self, exp, context):
        return exp.operator.do_evaluate(lambda: self.evaluate(exp.right, context))

    def visit_Literal(self, literal, context):
        return literal.value

    def visit_Identifier(self, identifier, context):
        if identifier.relative:
            subject = context.relative_value
        elif identifier.subject:
            subject = self.evaluate(identifier.subject, context)
        else:
            subject = context

        return subject.get(identifier.value, None)

    def visit_ObjectLiteral(self, object_literal, context):
        return dict(
            (key, self.evaluate(value, context))
            for key, value in object_literal.value.items()
        )

    def visit_ArrayLiteral(self, array_literal, context):
        return [self.evaluate(value, context) for value in array_literal.value]

    def visit_Function(self, func, context):
        try:
            function_func = self.config.functions[func.name]
        except KeyError:
            raise MissingFunctionError(
                'No function found with the name "{name}"'.format(name=func.name)
            )

        args = [self.evaluate(arg, context) for arg in func.args]
        return function_func(*args)

    def visit_Transform(self, transform, context):
        try:
            transform_func = self.config.transforms[transform.name]
        except KeyError:
            raise MissingTransformError(
                'No transform found with the name "{name}"'.format(name=transform.name)
            )

        args = [self.evaluate(arg, context) for arg in transform.args]
        return transform_func(self.evaluate(transform.subject, context), *args)

    def visit_FilterExpression(self, filter_expression, context):
        values = self.evaluate(filter_expression.subject, context)
        if filter_expression.relative:
            return [
                value
                for value in values
                if self.evaluate(
                    filter_expression.expression, context.with_relative(value)
                )
            ]
        else:
            filter_value = self.evaluate(filter_expression.expression, context)
            if filter_value is True:
                return values
            elif filter_value is False:
                return None
            else:
                try:
                    return values[filter_value]
                except (IndexError, KeyError):
                    return None

    def visit_ConditionalExpression(self, conditional, context):
        if self.evaluate(conditional.test, context):
            return self.evaluate(conditional.consequent, context)
        else:
            return self.evaluate(conditional.alternate, context)

    def generic_visit(self, expression, context):
        raise ValueError("Could not evaluate expression: " + repr(expression))
